;;; SSBL - a smaller serial bootloader for the ATmega8 and others
;;;
;;; This is an attempt to make a smaller serial bootloader by not
;;; requiring emulation of other protocols (stk500, avr910/avr109, etc)
;;; and putting more of the logic on the host side.
;;;
;;; Written by Wim Lewis <wiml@hhhh.org>
;;; 

#define __SFR_OFFSET 0
#include <avr/io.h>

;FOSC = 1000000			; factory-default internal osc setting
FOSC = 8000000			; set LFUSE to 0xE4 (8MHz internal osc)
BAUDRATE = 9600			; set this to whatever you want

.section .init0
	
.global normal_start
normal_start:	
		
.section .boot
bootloader_start:

#if defined(CHECK_PORT) && defined(CHECK_PIN)
	;; Check PORT D.4 (or other pin), jump to normal bootup if low.

/* This silly-looking set of macros is needed to get PORTfoo to expand
   correctly. If you really want to know why, check the CPP manual. */
#define PASTE(A,B) A ## B
#define DDR(port) PASTE(DDR,port)
#define PINS(port) PASTE(PIN,port)

#ifndef MINIMAL
	;; DDR D should be set to all-inputs by the reset code anyway
	clr r1
	out DDR(CHECK_PORT), r1		; set PORT D to all inputs
#endif
	sbis PINS(CHECK_PORT), CHECK_PIN	; read PORT D.4
	rjmp normal_start	; if zero, jump to normal boot address
#endif
	
	;; Set up the UART

	UBR = (FOSC / 8 / BAUDRATE) - 1
	ldi R24, lo8(UBR)
	out UBRRL, R24
	ldi R24, hi8(UBR)
	out UBRRH, R24

	ldi R24, (1<<TXC) | (1<<RXC) | (1<<U2X)
	out UCSRA, R24
	ldi R24, (1<<RXEN) | (1<<TXEN) ; Enable RX and TX
	out UCSRB, R24
	ldi R24, (1<<URSEL) | (1<<USBS) | (3<<UCSZ0) ; 8 data, 2 stop
	out UCSRC, R24

	;; Initialize the stack pointer
	;; To save a few bytes, we also jump back here on errors, instead
	;; of unwinding the stack. The T flag is set to indicate that we've
	;; done that (so we emit the error response instead of the
	;; initial ready response).
	clt
reset_stack_and_putch:	

	ldi R24, lo8(RAMEND)	;  SP <-- RAMEND
	ldi R25, hi8(RAMEND)
	out SPL, R24
	out SPH, R25

	brts putch_and_continue
	;; greeting msg
	ldi r16, 'o'
	rcall put_ch
	ldi r16, 'k'
	rcall put_ch
	
put_ok_and_continue:
	ldi r16, 13
putch_and_continue:
	;; wait for completion before sending ack byte
	in r17, SPMCR
	andi r17, (1<<SPMEN)
	brne putch_and_continue
	rcall put_ch
	;; fall through to command loop
		
	;; command loop
command_loop:	
	rcall get_ch

#ifndef MINIMAL
	cpi r16, 'p'
	brne not_idchar
	ldi r16, 's' 		; id char: little s for simple serial
	rjmp putch_and_continue
not_idchar:
#endif
	
	cpi r16, 'z'		; load Z register
	brne not_ldz
	rcall get_xbyte
	mov ZH, r16
	rcall get_xbyte
	mov ZL, r16
	rjmp put_ok_and_continue
not_ldz:
	
	cpi r16, 'R'		; read program memory
	brne not_lpm
	rcall get_xbyte
	mov r17, r16
2:	lpm r16, Z+
	rcall put_xbyte
	subi r17, 1
	brne 2b
	rjmp put_ok_and_continue
not_lpm:
	
	cpi r16, 'W'		; write program memory
	brne not_wpm
	rcall get_xbyte
	mov r17, r16
	ldi r18, (1<<SPMEN)
2:	rcall get_r0r1
	out SPMCR, r18
	spm
	adiw ZL,2
	subi r17, 1
	brne 2b
	rjmp put_ok_and_continue
not_wpm:
	
	cpi r16, 'P'		; load SPM control register
	brne not_spmcr
	rcall get_r0r1
	rcall get_xbyte
	out SPMCR, r16
	spm
	rjmp put_ok_and_continue
not_spmcr:
	
#ifndef MINIMAL
	cpi r16, 'L'
	brne not_lpmcr
	rcall get_xbyte
	out SPMCR, r16
	lpm r16, Z
	rcall put_xbyte
	rjmp put_ok_and_continue
not_lpmcr:	
#endif
	
	cpi r16, 'Q'		; query parameters
	brne not_query_parms
	ldi r16, 'z'
	rcall put_ch
	mov r16, ZH
	mov r17, ZL
	rcall put_xword
	ldi r16, 'b'
	rcall put_ch
	ldi r16, hi8(bootloader_start)
	ldi r17, lo8(bootloader_start)
	rcall put_xword
	rjmp put_ok_and_continue
not_query_parms:
	
	cpi r16, '@'		; call/jump to arbitrary address
	brne not_icall
	icall
	rjmp put_ok_and_continue
not_icall:
	
	;; no more command letters
	;; fall through to error case
	
err:
	ldi r16, '?'
	set
	rjmp reset_stack_and_putch

;;; utility

get_r0r1:
	rcall get_xbyte
	mov r0, r16
	rcall get_xbyte
	mov r1, r16
	ret
	
;;; UART routines

get_ch:
	sbis UCSRA, RXC 	; check for data ready
	rjmp get_ch		; if not, keep looping
	in r16, UDR		; else, retrieve data byte
	ret

get_xdigit:
	rcall get_ch		; retrieve a character in r16
	subi r16, '0'
	brlt err
	cpi r16, 10		; was it a decimal digit?
	brlt ok_xdigit		; if so, return it
	subi r16, ('A' - '0')
	brlt err
	cpi r16, 6		; okay, was it a letter?
	brlt ok_xletter
	subi r16, ('a' - 'A')
	brlt err
	cpi r16, 6
	brge err
ok_xletter:
	subi r16, -10
ok_xdigit:
	ret

get_xbyte:
	rcall get_xdigit
	swap r16
	mov r15, r16
	rcall get_xdigit
	or r16, r15
	ret

put_xword:
	rcall put_xbyte   	; print first byte from r16
	mov r16, r17		; second byte is in r17; fall through
	
put_xbyte:
	mov r15, r16
	swap r16
	rcall put_xdigit	; put high nybble
	mov r16, r15
	;rjmp put_xdigit        ; fall through to tail call of put_xdigit
		
put_xdigit:	
	andi r16, 0x0F		; mask off high byte; 0 ... +15
	subi r16, 10		; -10 ... +5
	brlt 1f
	subi r16, (- ('A' - '0' - 10)) ; 
1:	subi r16, (- (10 + '0'))
	;rjmp put_ch		; fall through to tail call to put_ch

put_ch:
	sbis UCSRA, UDRE	; check for space in tx fifo
	rjmp put_ch		; if not, keep looping
	out UDR, r16		; enqueue data byte
	ret			; return to caller


	
