#
# Host-side driver for ssbl.S
#

import fcntl, posix, re, string, sys, termios, time

error = 'ssbl.err'

class ssbl:
    avr = None
    pagesize = 64 # bytes
    qre = re.compile('([a-zG-Z])([0-9A-F]+)')
    oldattrs = None
    verbose = 0
    cur_z = None

    def __init__(self):
        self.query = { }

    # Opens the serial port, sets up baud rate &c
    def open(self, dev, baud=termios.B9600):
        t = termios
        self.avr = posix.open(dev, posix.O_RDWR | posix.O_NOCTTY)
        t.tcflush(self.avr, t.TCIOFLUSH)
        attrs = t.tcgetattr(self.avr)
        nattrs = [
            ( attrs[0] & ~(t.BRKINT | t.ICRNL | t.INPCK | t.IXON | t.IXOFF)) | t.ICRNL | t.ISTRIP,
            attrs[1] & ~(t.OPOST),
            ( attrs[2] & ~(t.CSIZE | t.PARENB | t.CRTSCTS) ) | t.CS8 | t.CLOCAL | t.CREAD,
            attrs[3] & ~(t.ECHO | t.ICANON | t.IEXTEN | t.ISIG),
            baud, baud,
            attrs[6][:]
            ] + attrs[7:]
        nattrs[6][t.VMIN] = 0
        nattrs[6][t.VTIME] = 20
        self.oldattrs = attrs
        t.tcsetattr(self.avr, t.TCSAFLUSH, nattrs)
        # fcntl.fcntl(self.avr, fcntl.F_SETFL, posix.O_NONBLOCK)

    # Close the serial port and restore tty settings
    def close(self):
        if self.avr is not None:
            if self.oldattrs is not None:
                termios.tcsetattr(self.avr, termios.TCSAFLUSH, self.oldattrs)
            posix.close(self.avr)
            self.avr = None

    def __del__(self):
        if self.avr is not None:
            posix.close(self.avr)
            self.avr = None

    def drain(self):
        while 1:
            s = posix.read(self.avr, 16)
            if len(s):
                print '...', repr(s)
            else:
                break

    # Perform one command, return the response
    def do(self, cmd, rbuflen=64):
        self.cur_z = None
        if self.verbose: print '-->', repr(cmd)
        n = posix.write(self.avr, cmd)
        if (n < len(cmd)):
            raise error, 'Failed to write cmd'
        s = posix.read(self.avr, rbuflen)
        if self.verbose: print '<--', repr(s)
        if s == '':
            raise error, 'Timed out waiting response'
        if s == '?':
            raise error, 'Command not supported ('+cmd+')'
        buf = s
        while '\r' not in buf and '\n' not in buf:
            s = posix.read(self.avr, rbuflen)
            if self.verbose: print '<--', repr(s)
            if s == '':
                raise error, 'Timed out awaiting rest of response'
            buf = buf + s
        return buf

    # Sets the Z register to a specified value, shortcircuiting if we
    # have some way of knowing that it already has that value
    def setz(self, addr):
        if addr != self.cur_z:
            self.do('z%04X' % (addr,))
            self.cur_z = addr

    # query (useful as a ping)
    def do_q(self):
	r = self.do('Q')
        p = 0
	while 1:
            m = self.qre.match(r, p)
            if m is None:
                break
            (k, v) = m.groups()
            p = m.end()
            if k == 'z':
                self.cur_z = int(v, 16)
            else:
                self.query[k] = v
    #

    # Write one page of data at the specified address
    def pagewrite(self, addr, data):
        setz = 'z%04X' % (addr,)
        self.do(setz)
        dlen = len(data)
        if dlen == 0:
            return
        if (dlen % 2) != 0:
            raise error, 'Must write a whole number of words'
        self.do('P000003')  # erase page
        self.do('P000011')  # enable read-while-write (why necc.?)
        self.do('Q')
        words = dlen / 2
        self.do('W%02X' % (words,) +
                string.join(map(lambda x: '%02X' % (ord(x),), data),''))
        self.do('Q')
        self.do(setz)
        self.do('P000005')  # write page
        self.do('P000011')  # enable read-while-write (why necc.?)
        self.do('Q')

    # Read a block of memory (at most 256 bytes)
    def readmem(self, addr, nbytes):
        self.setz(addr)
        b = self.do('R%02X' % (nbytes,)).strip()
        buf = ''
        for i in range(0, len(b), 2):
            buf = buf + chr(int(b[i:i+2],16))
        if len(buf) == nbytes:
            self.cur_z = addr + nbytes
        return buf


# This ihex-parsing routine is adapted from uavrp.py by Jeff Epler
def read_ihex(f):
    memory = {}
    for line in f:
        line = line.strip()
        if not line.startswith(":"): continue
        bytecount = int(line[1:3], 16)
        address = int(line[3:7], 16)
        rectype = int(line[7:9], 16)
        data = line[9:-2]
        checksum = int(line[-2:], 16)
        if 2 * bytecount != len(data):
            raise error, "Bad byte count (%dd/%dx) in ihex record" % (bytecount,len(data))
        check = sum(map(lambda i: int(line[i:i+2],16), range(1, len(line), 2)))
        if check % 256 != 0:
            raise error, "bad checksum in ihex record (0x%x)" % check
        if rectype > 1:
            raise error, "Unsupported record type 0x%02x" % rectype
        if rectype == 1:
            break
        for i in range(address, address + len(data)/2):
            j = 2 * (i - address)
            memory[i] = int(data[j:j+2], 16)
    l = [-1] * (max(memory.keys()) + 1)
    for k, v in memory.items(): l[k] = v
    return l

# Given an 'avr' instance and a memory map, write it to the flash
def writeall(avr, memmap, precheck=0, dots=1):
    if dots:
        sys.stdout.write("writing")
        t0 = time.time()
    pagecount = 0
    for addr in range(0, len(memmap), avr.pagesize):
        page = memmap[addr : addr+avr.pagesize]
        if page == ( [-1] * len(page) ):
            continue
        if len(page) < avr.pagesize:
            page = page + ( [-1] * (avr.pagesize * len(page)) )
        if precheck or (-1 in page):
            old = avr.readmem(addr, avr.pagesize)
        buf = ''
        for subaddr in range(0, avr.pagesize):
            if page[subaddr] != -1:
                buf = buf + chr(page[subaddr])
            else:
                buf = buf + old[subaddr]
        if old == buf:
            if dots:
                sys.stdout.write('.')
        else:
            avr.pagewrite(addr, buf)
            pagecount = pagecount + 1
            if dots:
                sys.stdout.write('#')
        if dots:
            sys.stdout.flush()
    if dots:
        totalpagecount = ( len(memmap) + avr.pagesize - 1) / avr.pagesize
        sys.stdout.write(" (wrote %d of %d pages, %.1fsec)\n" % (pagecount, totalpagecount, time.time() - t0))

# Verify an AVR's flash against a memory map
def verify(avr, memmap, dots=1):
    if dots:
        sys.stdout.write(" verify")
        t0 = time.time()
    bytecount = 0
    for addr in range(0, len(memmap), avr.pagesize):
        page = memmap[addr : addr+avr.pagesize]
        if page == ( [-1] * len(page) ):
            continue
        if len(page) < avr.pagesize:
            page = page + ( [-1] * (avr.pagesize * len(page)) )
        buf = avr.readmem(addr, avr.pagesize)
        bytecount = bytecount + avr.pagesize
        for subaddr in range(0, avr.pagesize):
            if page[subaddr] != -1:
                rd = ord(buf[subaddr])
                if rd != page[subaddr]:
                    print "Mismatch at address %04x: read %02x, want %02x" % (addr+subaddr, rd, page[subaddr])
        if dots:
            sys.stdout.write('.')
            sys.stdout.flush()
    if dots:
        sys.stdout.write(" (%dB / %.1fsec)\n" % ( bytecount, (time.time() - t0) ))

#  Here's an example of using the bootloader to store an IHEX file
def store_ihex(filename, a):
    M = read_ihex(open(filename))
    a.do_q()
    if a.query.has_key('b'):
        begin = int(a.query['b'],16)
        if len(M) >= begin:
            print 'Failure: memory to write (0x%04x) would stomp bootloader (0x%04x)' % (len(M), begin)
            a.close()
            return
    oldverbose = a.verbose
    a.verbose = 0
    writeall(a, M, dots=1, precheck=1)
    verify(a, M, dots=1)
    a.verbose = oldverbose

#
# If ssbl.py is not loaded as a module, provide a super-simple command
# line interface
#
if __name__ == '__main__':
    if len(sys.argv) != 3:
        print 'simple command-line usage: python ssbl.py /dev/foo bar.hex'
        print 'for more complex situations, load ssbl.py as a module'
        sys.exit(1)
    print 'Opening device %s, loading ihex file %s...' % (sys.argv[1],sys.argv[2])
    avr = ssbl()
    avr.verbose = 1
    avr.open(sys.argv[1])
    avr.drain()
    store_ihex(sys.argv[2], avr)
    avr.close()
